@@ -7,15 +7,16 @@ import hashlib
 from twisted.trial import unittest
 
 from twisted.python.filepath import FilePath
+from twisted.internet.defer import succeed
 from twisted.internet.task import Clock
-from twisted.web.test import _util
 from twisted.web import http, server, resource
 
 from webmagic.filecache import FileCache
-from webmagic.fakes import DummyChannel, DummyRequest
+from webmagic.fakes import DummyChannel, DummyRequest, DummyTCPTransport
 from webmagic.untwist import (
 	CookieInstaller, BetterResource, RedirectingResource, HelpfulNoResource,
-	_CSSCacheEntry, BetterFile, ResponseCacheOptions, setCachingHeadersOnRequest
+	_CSSCacheEntry, BetterFile, ResponseCacheOptions,
+	setCachingHeadersOnRequest, BetterSite
 )
 
 
@@ -182,6 +183,17 @@ class NonLeafPlainResource(resource.Resource):
 
 
 
+class DynamicBetterResource(BetterResource):
+	path = None
+	childResource = BetterResource()
+
+	def getChild(self, path, request):
+		self.path = path
+		self.request = request
+		return self.childResource
+
+
+
 class BetterResourceTests(unittest.TestCase):
 
 	def _makeSite(self, r):
@@ -356,6 +368,17 @@ class BetterResourceTests(unittest.TestCase):
 		self.assertTrue(isinstance(res, HelpfulNoResource), res)
 
 
+	def test_getChildCalledForNonexistentChild(self):
+		req = DummyRequest([''])
+		r = DynamicBetterResource()
+		uri = 'hello'
+
+		child = r.getChildWithDefault(uri, req)
+		self.assertIdentical(child, r.childResource)
+		self.assertEqual(r.path, uri)
+		self.assertIdentical(r.request, req)
+
+
 	# Right now, the behavior is to 404 if there are any extra slashes,
 	# except for the root Resource, which strange accepts 1 extra slash.
 
@@ -404,6 +427,21 @@ class CSSCacheEntryTests(unittest.TestCase):
 
 
 
+def _render(resource, request):
+	result = resource.render(request)
+	if isinstance(result, str):
+		request.write(result)
+		request.finish()
+		return succeed(None)
+	elif result is server.NOT_DONE_YET:
+		if request.finished:
+			return succeed(None)
+		else:
+			return request.notifyFinish()
+	else:
+		raise ValueError("Unexpected return value: %r" % (result,))
+
+
 class BetterFileTests(unittest.TestCase):
 
 	def _makeDummyRequest(self, postpath, path, site):
@@ -418,7 +456,7 @@ class BetterFileTests(unittest.TestCase):
 	def _requestPostpathAndRender(self, baseResource, postpath, path=None, site=None):
 		request = self._makeDummyRequest(postpath, path, site)
 		child = resource.getChildForRequest(baseResource, request)
-		d = _util._render(child, request)
+		d = _render(child, request)
 		d.addCallback(lambda _: (request, child))
 		return d
 
@@ -678,3 +716,41 @@ class TestsetCachingHeadersOnRequest(unittest.TestCase):
 			'Date': ['Thu, 01 Jan 1970 00:00:00 GMT'],
 			'Expires': ['-1']},
 		dict(request.responseHeaders.getAllRawHeaders()))
+
+
+
+class BetterSiteTests(unittest.TestCase):
+
+	def test_noDelayTrue(self):
+		"""
+		The default options for BetterSite cause NO_DELAY to be set to True
+		on new connections.
+		"""
+		br = BetterResource()
+		bs = BetterSite(br)
+		channel = bs.buildProtocol(None)
+		transport = DummyTCPTransport()
+		channel.makeConnection(transport)
+		self.assertTrue(transport.everCalledSetTcpNoDelay())
+		self.assertTrue(transport.getTcpNoDelay())
+
+		# Lose the connection to clear HTTPChannel.timeOut, so we don't
+		# have a dirty reactor.
+		channel.connectionLost(None)
+
+
+	def test_noDelayFalse(self):
+		"""
+		If noDelay=False is passed to BetterSite, it does not set NO_DELAY
+		on new connections, leaving the default option.
+		"""
+		br = BetterResource()
+		bs = BetterSite(br, noDelay=False)
+		channel = bs.buildProtocol(None)
+		transport = DummyTCPTransport()
+		channel.makeConnection(transport)
+		self.assertFalse(transport.everCalledSetTcpNoDelay())
+
+		# Lose the connection to clear HTTPChannel.timeOut, so we don't
+		# have a dirty reactor.
+		channel.connectionLost(None)
