@@ -10,6 +10,12 @@ import time
 from functools import partial
 
 from twisted.web import resource, static, server
+try:
+	# Twisted >= 9.0
+	from twisted.web.resource import ErrorPage
+except ImportError:
+	from twisted.web.error import ErrorPage
+
 from twisted.web.http import HTTPChannel, datetimeToString
 from twisted.python import context, log
 
@@ -18,6 +24,7 @@ from zope.interface import implements
 from webmagic.transforms import md5hexdigest
 from webmagic.pathmanip import ICacheBreaker
 from webmagic.cssfixer import fixUrls
+from webmagic.safe_headers import setRawHeadersSafely
 
 _postImportVars = vars().keys()
 
@@ -152,7 +159,7 @@ def setNoCacheNoStoreHeaders(request):
 	setRawHeaders('expires', ['-1'])
 
 
-class HelpfulNoResource(resource.ErrorPage):
+class HelpfulNoResource(ErrorPage):
 	template = """\
 <!doctype html>
 <html>
@@ -167,13 +174,13 @@ class HelpfulNoResource(resource.ErrorPage):
 </html>"""
 
 	def __init__(self, message='Page not found. <a href="/">See the index?</a>'):
-		resource.ErrorPage.__init__(
+		ErrorPage.__init__(
 			self, 404, "404 Not Found", message)
 
 
 	def render(self, request):
 		setDefaultHeadersOnRequest(request)
-		return resource.Resource.render(self, request)
+		return ErrorPage.render(self, request)
 
 
 
@@ -204,23 +211,26 @@ class RedirectingResource(resource.Resource):
 	def render(self, request):
 		setDefaultHeadersOnRequest(request)
 		request.setResponseCode(self._code)
-		# twisted.web z9trunk protects against response-splitting, so we don't
-		# need to do anything to the Location header.
-		# Also, this is a relative redirect; non-standard, but all browsers accept it.
-		request.responseHeaders.setRawHeaders('location', [self._location])
+		# This is a relative redirect, so it is non-standard, but all
+		# browsers accept it.
+		setRawHeadersSafely(
+			request.responseHeaders, 'location', [self._location])
 		return self.template % {'escaped': cgi.escape(self._location)}
 
 
 
 class BetterResource(resource.Resource):
 	"""
-	A L{resource.Resource} with two improvements:
+	A L{resource.Resource} with several improvements:
 
 	*	/page and /page/ are forced to be the same thing (/page is
 		redirected to /page/)
 
 	*	Additional response headers are set for security reasons.
 
+	*	Cache-related headers are removed if an exception was raised from
+		render().
+
 	Implementation notes:
 
 	By default, twisted.web Resources with `isLeaf = True`:
@@ -249,6 +259,32 @@ class BetterResource(resource.Resource):
 			raise
 
 
+	def getChild(self, path, request):
+		"""
+		Works like L{resource.Resource.getChild}:
+
+		Retrieve a 'child' resource from me.
+
+		Implement this to create dynamic resource generation -- resources which
+		are always available may be registered with self.putChild().
+
+		BetterResource-specific:
+
+		All that's implemented by default beyond a L{resource.Resource}
+		is to replace the 404 page with a slightly better one.
+
+		Note that what L{getChildWithDefault} enforces on resources
+		with trailing crud, and what it implements for redirecting
+		paths without slashes, is not implemented here for dynamic
+		resources.  Subclasses should implement this behavior themselves
+		if it is desired.
+		"""
+		if self._debugGetChild:
+			log.msg("BetterResource: Returning 404 "
+				"because no suitable resource")
+		return HelpfulNoResource()
+
+
 	def getChildWithDefault(self, path, request):
 		"""
 		We want to serve the same 404 page for these two cases:
@@ -265,12 +301,10 @@ class BetterResource(resource.Resource):
 			log.msg("BetterResource: prepath=%r postpath=%r uri=%r" % (
 				request.prepath, request.postpath, request.uri))
 
-		# 404 requests for which there is no suitable Resource
+		# Like resource.Resource, give getChild a chance to respond
+		# with a dynamic resource for paths that aren't in self.children
 		if not path in self.children:
-			if self._debugGetChild:
-				log.msg("BetterResource: Returning 404 "
-					"because no suitable resource")
-			return HelpfulNoResource()
+			return self.getChild(path, request)
 
 		# 404 requests that have extra crud
 		if self.children[path].isLeaf and request.postpath not in ([], ['']):
@@ -307,11 +341,40 @@ def loadCompatibleMimeTypes():
 	# Read from Python's built-in mimetypes, but don't load any mimetypes
 	# from disk.
 	contentTypes = static.loadMimeTypes(mimetype_locations=())
-	# Send the mimetypes that Google sends. These were captured on 2010-06-09.
 	contentTypes.update({
+		# Use text/javascript instead of application/javascript for
+		# compatibility reasons.
 		'.js': 'text/javascript',
 		'.ico': 'image/x-icon',
 		'.log': 'text/plain',
+		'.rst': 'text/plain',
+
+		# Office 2007 formats
+		# http://technet.microsoft.com/en-us/library/ee309278%28office.12%29.aspx
+		'.docx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
+		'.docm': 'application/vnd.ms-word.document.macroEnabled.12',
+		'.dotx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.template',
+		'.dotm': 'application/vnd.ms-word.template.macroEnabled.12',
+		'.xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
+		'.xlsm': 'application/vnd.ms-excel.sheet.macroEnabled.12',
+		'.xltx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.template',
+		'.xltm': 'application/vnd.ms-excel.template.macroEnabled.12',
+		'.xlsb': 'application/vnd.ms-excel.sheet.binary.macroEnabled.12',
+		'.xlam': 'application/vnd.ms-excel.addin.macroEnabled.12',
+		'.pptx': 'application/vnd.openxmlformats-officedocument.presentationml.presentation',
+		'.pptm': 'application/vnd.ms-powerpoint.presentation.macroEnabled.12',
+		'.ppsx': 'application/vnd.openxmlformats-officedocument.presentationml.slideshow',
+		'.ppsm': 'application/vnd.ms-powerpoint.slideshow.macroEnabled.12',
+		'.potx': 'application/vnd.openxmlformats-officedocument.presentationml.template',
+		'.potm': 'application/vnd.ms-powerpoint.template.macroEnabled.12',
+		'.ppam': 'application/vnd.ms-powerpoint.addin.macroEnabled.12',
+		'.sldx': 'application/vnd.openxmlformats-officedocument.presentationml.slide',
+		'.sldm': 'application/vnd.ms-powerpoint.slide.macroEnabled.12',
+		'.one': 'application/onenote',
+		'.onetoc2': 'application/onenote',
+		'.onetmp': 'application/onenote',
+		'.onepkg': 'application/onenote',
+		'.thmx': 'application/vnd.ms-officetheme',
 	})
 	return contentTypes
 
@@ -583,35 +646,60 @@ class BetterFile(static.File):
 
 
 
-class ConnectionTrackingHTTPChannel(HTTPChannel):
+class _BetterHTTPChannel(HTTPChannel):
+	def connectionMade(self, *args, **kwargs):
+		HTTPChannel.connectionMade(self, *args, **kwargs)
+		if self.site._setNoDelayOnConnect and \
+		hasattr(self.transport, 'setTcpNoDelay'):
+			self.transport.setTcpNoDelay(True)
+
+
+
+class BetterSite(server.Site):
+	"""
+	A L{server.Site} with a few modifications:
+
+	*	Default idle (not serving a request) keep-alive timeout of 75 seconds
+		instead of 12 hours.
+
+	*	Sets TCP_NODELAY on all connections (unless disabled with
+		noDelay=False).
+	"""
+	protocol = _BetterHTTPChannel
+
+	def __init__(self, resource, logPath=None, timeout=75, noDelay=True):
+		server.Site.__init__(self, resource, logPath, timeout)
+		self._setNoDelayOnConnect = noDelay
+
+
+
+class ConnectionTrackingHTTPChannel(_BetterHTTPChannel):
 	"""
 	An L{HTTPChannel} that tells the factory about all connection
 	activity.
 	"""
-	__slots__ = ()
-
 	def __init__(self, *args, **kwargs):
-		HTTPChannel.__init__(self, *args, **kwargs)
+		_BetterHTTPChannel.__init__(self, *args, **kwargs)
 
 
 	def connectionMade(self, *args, **kwargs):
-		HTTPChannel.connectionMade(self, *args, **kwargs)
+		_BetterHTTPChannel.connectionMade(self, *args, **kwargs)
 		log.msg('Connection made: %r' % (self,))
 		self.factory.connections.add(self)
 
 
 	def connectionLost(self, *args, **kwargs):
-		HTTPChannel.connectionLost(self, *args, **kwargs)
+		_BetterHTTPChannel.connectionLost(self, *args, **kwargs)
 		log.msg('Connection lost: %r' % (self,))
 		self.factory.connections.remove(self)
 
 
 
-class ConnectionTrackingSite(server.Site):
+class ConnectionTrackingSite(BetterSite):
 	protocol = ConnectionTrackingHTTPChannel
 
 	def __init__(self, *args, **kwargs):
-		server.Site.__init__(self, *args, **kwargs)
+		BetterSite.__init__(self, *args, **kwargs)
 		self.connections = set()
 
 
